/***************************************************************************
 *   Copyright (C) 2009 by Andreas Fink <finkandreas@web.de>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// KDE includes
#include <Plasma/Meter>
#include <Plasma/SignalPlotter>


// Qt includes
#include <QGraphicsGridLayout>
#include <QFile>
#include <QTextStream>

// local includes
#include "scriptparser.h"
#include "yasp-scripted.h"
#include "frame.h"
#include "label.h"
#include "iconstack.h"

ScriptParser::ScriptParser( Yasp* pYasp )
    : m_pYasp(pYasp)
{
}


void ScriptParser::run()
{
  m_nInterval = 5000;
  m_TitleFont = m_ValueFont = Plasma::Theme::defaultTheme()->font(Plasma::Theme::DesktopFont);
  m_TitleColor = m_ValueColor = Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);
  m_TitleAlignment = m_ValueAlignment = Qt::AlignCenter;
  m_TitleShadow = Frame::Sunken;
  m_Row = 0;
  m_RegisteredSensors.clear();
  m_GuiElements.clear();
  m_bParsingSensors = true;
  m_bReparseRequested = false;

  QFile file( m_pYasp->ScriptPath() );
  file.open( QIODevice::ReadOnly );
  QTextStream s( &file );
  QString sLine;
  bool bContinue = true;
  do {
    sLine = s.readLine().trimmed();
    while ( sLine.endsWith("\\") )
      sLine.replace( QRegExp("\\\\$"), s.readLine().trimmed() );
//    qDebug() << "Read Line: " << sLine;
    if ( sLine.startsWith("sensor") )
      bContinue = AddSensor( sLine.section(' ', 1).trimmed() );
    else if ( sLine.startsWith("default") ) {
      m_GuiElements << sLine;
      bContinue = SetDefaults( sLine.section(' ', 1).trimmed() );
    }
    else if ( sLine.startsWith("title") || sLine.startsWith("value") ||
              sLine.startsWith("meter") || sLine.startsWith("plotter") ||
              sLine.startsWith("text")  || sLine.startsWith("icons") )
    {
      m_GuiElements << sLine;
    }
  } while( !sLine.isNull() && bContinue );

  if ( m_bReparseRequested )
    return;

  WaitForDependencies(m_RegisteredSensors);
}


void ScriptParser::ParseGuiElements()
{
  m_bParsingSensors = false;
  bool bContinue = true;
  foreach( QString sLine, m_GuiElements ) {
    if ( sLine.startsWith("default") )
      bContinue = SetDefaults( sLine.section(' ', 1).trimmed() );
    else if ( sLine.startsWith("title") )
      bContinue = AddTitle( sLine.section(' ', 1).trimmed() );
    else if ( sLine.startsWith("value") )
      bContinue = AddValue( sLine.section(' ', 1).trimmed() );
    else if ( sLine.startsWith("meter") )
      bContinue = AddMeter( sLine.section(' ', 1).trimmed() );
    else if ( sLine.startsWith("plotter") )
      bContinue = AddSignalPlotter( sLine.section(' ', 1).trimmed() );
    else if ( sLine.startsWith("text") )
      bContinue = AddText( sLine.section(' ', 1).trimmed() );
    else if ( sLine.startsWith("icons") )
      bContinue = AddIcons( sLine.section(' ', 1).trimmed() );

    if ( bContinue == false ) {
      m_pYasp->ParsingError( "Broken dependency in the line: " + sLine );
      return;
    }
  }
}


bool ScriptParser::ParseKeyValue( const QString& sConfig, QString& sKey, QString& sValue, int& nStartIdx ) const
{
  int nKeyEndIdx = sConfig.indexOf( '=', nStartIdx );
  if ( -1 == nKeyEndIdx )
    return false;

  sKey = sConfig.mid( nStartIdx, nKeyEndIdx-nStartIdx ).trimmed();
  QChar sep = sConfig.at( ++nKeyEndIdx );
  int nValueEndIdx = sConfig.indexOf( sep, ++nKeyEndIdx );
  sValue = sConfig.mid( nKeyEndIdx, nValueEndIdx-nKeyEndIdx );
  nStartIdx = nValueEndIdx+1;
  return true;
}


bool ScriptParser::AddTitle( const QString& sConfig )
{
  QString sText = "";
  QFont Font = m_TitleFont;
  QColor Color = m_TitleColor;
  Frame::Shadow Shadow = m_TitleShadow;
  Qt::Alignment Align = m_TitleAlignment;

  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
    if ( "text" == sKey )
      sText = sValue;
    if ( "font" == sKey )
      Font.fromString( sValue );
    if ( "color" == sKey )
      Color = QColor( sValue );
    if ( "shadow" == sKey )
      Shadow = sValue.toLower()=="sunken" ? Frame::Sunken : sValue.toLower()=="raised" ? Frame::Raised : Frame::Plain;
    if ( "alignment" == sKey )
      Align = StringToAlignment(sValue);
  }

  Frame* f = new Frame( m_pYasp );
  f->setText( ReplaceFormatChars(sText) );
  f->updateAppearance( Color, Font, Shadow, Align );
  m_pYasp->layout()->addItem( f, m_Row++, 0, 1, 2 );

  return true;
}


bool ScriptParser::AddSensor( const QString& sConfig )
{
  QString sType;
  QString sCmd;
  int nInterval = m_nInterval;
  QString sMathOperation;
  QString sSensorName;
  QStringList UsedSensors;

  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
    if ( "type" == sKey )
      sType = sValue;
    if ( "cmd" == sKey )
      sCmd = sValue;
    if ( "interval" == sKey )
      nInterval = sValue == "single" ? 0 : sValue.toInt() < 500 ? 500 : sValue.toInt();
    if ( "math" == sKey )
      sMathOperation = sValue;
    if ( "name" == sKey )
      sSensorName = sValue;
    if ( "use" == sKey )
      UsedSensors << sValue;
  }

  if ( sSensorName.isEmpty() )
    return false;

  if ( false == WaitForDependencies(UsedSensors) )
    return false;

  if ( "engine" == sType ) {
    QStringList EngineDesc;
    int left = sCmd.indexOf(':'), right = sCmd.lastIndexOf(':');
    if ( left == right ) {
      m_pYasp->ParsingError(QString("Error parsing sensor (less than 2 colons found): ")+sCmd);
    }
    else {
      EngineDesc << sCmd.left(left) << sCmd.mid(left+1, right-left-1) << sCmd.mid(right+1);
      SourceIntervalPair sip = { EngineDesc[0], EngineDesc[1], EngineDesc[2], nInterval };
      SensorCmdPair scp = { sSensorName, EngineDesc[2], sMathOperation, UsedSensors };
      emit EngineSensorAdded( sip, scp );
    }
  }

  if ( ("program" == sType && !sCmd.isEmpty()) || ("math" == sType && !sMathOperation.isEmpty()) ) {
    emit ProgramSensorAdded( (SensorCmdPair){sSensorName, sCmd, sMathOperation, UsedSensors}, nInterval );
  }

  m_RegisteredSensors << sSensorName;

  return true;
}


bool ScriptParser::AddValue( const QString& sConfig )
{
  QString sValueKey;
  QFont Font = m_ValueFont;
  QColor Color = m_ValueColor;
  Qt::Alignment Align = m_ValueAlignment;
  QString sIcon;
  QStringList UsedSensors;
  QString sFormat( "$1" );
  int nInterval = m_nInterval;
  int nIconSize = 0;
  QString sIconClick, sKeyClick, sValueClick;

  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
      if ( "key" == sKey )
      sValueKey = sValue;
    if ( "font" == sKey )
      Font.fromString( sValue );
    if ( "color" == sKey )
      Color = QColor( sValue );
    if ( "alignment" == sKey )
      Align = StringToAlignment( sValue );
    if ( "use" == sKey )
      UsedSensors << sValue;
    if ( "format" == sKey )
      sFormat = sValue;
    if ( "iconsize" == sKey )
      nIconSize = sValue.toInt();
    if ( "icon" == sKey )
      sIcon = sValue;
    if ( "icon_click" == sKey )
      sIconClick = sValue;
    if ( "key_click" == sKey )
      sKeyClick = sValue;
    if ( "value_click" == sKey )
      sValueClick = sValue;
  }

  if ( false == WaitForDependencies(UsedSensors) )
    return false;

  if ( ! sValueKey.isEmpty() ) {
    QPixmap p;
    if ( !sIcon.isEmpty() )
      p = KIconLoader::global()->loadIcon(sIcon, KIconLoader::NoGroup, nIconSize);
    Label* l = new Label( p, m_pYasp );
    l->setText( ReplaceFormatChars(sValueKey) );
    l->updateAppearance( Color, Font, Align );
    l->setIconClick( sIconClick );
    l->setTextClick( sKeyClick );
    m_pYasp->layout()->addItem( l, m_Row, 0, 1, 1 );
  }

  if ( ! UsedSensors.isEmpty() ) {
    Label* l = new Label( QPixmap(), m_pYasp );
    l->updateAppearance( Color, Font, Align );
    l->setTextClick( sValueClick );
    emit LabelAdded( (SensorLabelPair){l, UsedSensors, ReplaceFormatChars(sFormat)}, nInterval );
    m_pYasp->layout()->addItem( l, m_Row, 1, 1, 1 );
  }

  m_Row++;
  return true;
}


bool ScriptParser::SetDefaults( const QString& sConfig ) {
  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
    if( "title:color" == sKey )
      m_TitleColor = QColor( sValue );
    if ( "title:font" == sKey )
      m_TitleFont.fromString( sValue );
    if ( "title:alignment" == sKey )
      m_TitleAlignment = StringToAlignment( sValue );
    if ( "title:shadow" == sKey )
      m_TitleShadow = sValue.toLower()=="sunken" ? Frame::Sunken :
                      sValue.toLower()=="raised" ? Frame::Raised : Frame::Plain;
    if ( "value:color" == sKey )
      m_ValueColor = QColor( sValue );
    if ( "value:font" == sKey )
      m_ValueFont.fromString( sValue );
    if ( "value:alignment" == sKey )
      m_ValueAlignment = StringToAlignment( sValue );
    if ( "interval" == sKey )
      m_nInterval = sValue.toInt();
  }

  return true;
}


bool ScriptParser::AddMeter( const QString& sConfig )
{
  QHash<int, QFont> Fonts;
  QHash<int, QColor> Colors;
  QHash<int, Qt::Alignment> Alignments;
  QStringList UsedSensors;
  QHash<int, QString> Labels;
  int nInterval = m_nInterval;
  QString sMeterMin="0", sMeterMax="100", sMeterValue="0";
  QString sSvg = "widgets/bar_meter_horizontal";
  int min_height = Plasma::Meter().preferredHeight();

  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
    int nIdx = sKey.split(':').value(1, "").toInt();
    if ( sKey.startsWith("label") )
      Labels[nIdx] = ReplaceFormatChars(sValue);
    if ( sKey.startsWith("font") )
      Fonts[nIdx].fromString( sValue );
    if ( sKey.startsWith("color") )
      Colors[nIdx] = QColor( sValue );
    if ( sKey.startsWith("alignment") )
      Alignments[nIdx] = StringToAlignment(sValue);
    if ( "use" == sKey )
      UsedSensors << sValue;
    if ( "min" == sKey )
      sMeterMin = sValue;
    if ( "max" == sKey )
      sMeterMax = sValue;
    if ( "value" == sKey )
      sMeterValue = sValue;
    if ( "svg" == sKey )
      sSvg = sValue;
    if ( "interval" == sKey )
      nInterval = sValue == "single" ? 0 : sValue.toInt() < 500 ? 500 : sValue.toInt();
    if ( "min_height" == sKey )
      min_height = sValue.toInt();
  }

  if ( false == WaitForDependencies(UsedSensors) )
    return false;

  Plasma::Meter* m = new Plasma::Meter( m_pYasp );
  m->setMeterType( Plasma::Meter::BarMeterHorizontal );
  m->setSvg( sSvg );
  for ( QHash<int, QString>::const_iterator it = Labels.begin(); it != Labels.end(); ++it ) {
    m->setLabelAlignment( it.key(), m_ValueAlignment );
    m->setLabelColor( it.key(), m_ValueColor );
    m->setLabelFont( it.key(), m_ValueFont );
  }
  for ( QHash<int, QFont>::const_iterator it = Fonts.begin(); it != Fonts.end(); ++it )
    m->setLabelFont( it.key(), it.value() );
  for ( QHash<int, Qt::Alignment>::const_iterator it = Alignments.begin(); it != Alignments.end(); ++it )
    m->setLabelAlignment( it.key(), it.value() );
  for ( QHash<int, QColor>::const_iterator it = Colors.begin(); it != Colors.end(); ++it )
    m->setLabelColor( it.key(), it.value() );
  m->setPreferredHeight(min_height);
  m->setSizePolicy( QSizePolicy::Expanding, QSizePolicy::MinimumExpanding );

  emit MeterAdded( (SensorMeterPair){m, UsedSensors, sMeterMin, sMeterMax, sMeterValue, Labels}, nInterval );
  m_pYasp->layout()->addItem( m, m_Row++, 0, 1, 2 );

  return true;
}


bool ScriptParser::AddSignalPlotter( const QString& sConfig )
{
  QStringList Plots;
  QList< QColor > Colors;
  QStringList UsedSensors;
  int nHeight = 50;
  QString min, max;
  int nInterval = m_nInterval;
  QColor bgcolor;
  QColor gridcolor = Qt::white;
  int bgalpha = 255;
  int gridalpha = 255;
  QString sClick;
  bool bStack = false;

  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
    if ( "plot" == sKey )
      Plots << sValue;
    if ( "color" == sKey )
      Colors << QColor( sValue );
    if ( "min" == sKey )
      min = sValue;
    if ( "max" == sKey)
      max = sValue;
    if ( "height" == sKey )
      nHeight = sValue.toInt() == 0 ? 50 : sValue.toInt();
    if ( "use" == sKey )
      UsedSensors << sValue;
    if ( "backgroundcolor" == sKey )
      bgcolor = QColor( sValue );
    if ( "backgroundalpha" == sKey )
      bgalpha = sValue.toInt();
    if ( "gridcolor" == sKey )
      gridcolor = QColor( sValue );
    if ( "gridalpha" == sKey )
      gridalpha = sValue.toInt();
    if ( "interval" == sKey )
      nInterval = ( sValue == "single" ? 0 : ( sValue.toInt() < 500 ? 500 : sValue.toInt() ) );
    if( "stack" == sKey )
      bStack = ( sValue == "true" ? true : false );
  }

  if ( false == WaitForDependencies(UsedSensors) )
    return false;

  gridcolor.setAlpha( gridalpha );
  Plasma::SignalPlotter* p = new Plasma::SignalPlotter( m_pYasp );
  for ( int i=0; i<Plots.size(); ++i )
    p->addPlot( Colors.value(i, Qt::white) );
  if ( min.isEmpty() && max.isEmpty() )
    p->setUseAutoRange( true );
  else {
    p->setUseAutoRange( false );
    p->setVerticalRange( min.toDouble(), max.toDouble() );
  }
  p->setShowLabels( false );
  p->setShowVerticalLines( true );
  p->setVerticalLinesColor( gridcolor );
  p->setVerticalLinesDistance( 20 );
  p->setShowHorizontalLines( true );
  p->setHorizontalLinesCount( nHeight/20 );
  p->setHorizontalLinesColor( gridcolor );
  p->setThinFrame( false );
  p->setShowTopBar( false );
  p->setStackPlots( bStack );
  qDebug() << p->stackPlots();
  p->setPreferredHeight( nHeight );
  p->setSizePolicy( QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding );
  if ( bgcolor.isValid() ) {
    p->setSvgBackground( "" );
    bgcolor.setAlpha( bgalpha );
    p->setBackgroundColor( bgcolor );
  }

  emit PlotterAdded( (SensorPlotterPair){p, UsedSensors, Plots}, nInterval );
  m_pYasp->layout()->addItem( p, m_Row++, 0, 1, 2 );

  return true;
}


bool ScriptParser::AddText( const QString& sConfig )
{
  QString sText;
  QFont Font = m_ValueFont;
  QColor Color = m_ValueColor;
  Qt::Alignment Align = m_ValueAlignment;
  QStringList UsedSensors;
  QString sFormat = "$1";
  int nInterval = m_nInterval;
  QString sClick;

  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
      if ( "text" == sKey )
      sText = sValue;
    if ( "font" == sKey )
      Font.fromString( sValue );
    if ( "color" == sKey )
      Color = QColor( sValue );
    if ( "alignment" == sKey )
      Align = StringToAlignment( sValue );
    if ( "use" == sKey )
      UsedSensors << sValue;
    if ( "format" == sKey )
      sFormat = sValue;
    if ( "click" == sKey )
      sClick = sValue;
  }

  if ( false == WaitForDependencies(UsedSensors) )
    return false;

  Label* l = new Label( QPixmap(), m_pYasp );
  l->updateAppearance( Color, Font, Align );
  l->setTextClick( sClick );
  emit TextAdded( (SensorTextPair){l, UsedSensors, ReplaceFormatChars(sFormat)}, nInterval );
  m_pYasp->layout()->addItem( l, m_Row, 0, 1, 2 );

  m_Row++;
  return true;
}


bool ScriptParser::AddIcons( const QString& sConfig )
{
  int nSize=16;
  QStringList Icons;
  QStringList Clicks;

  int nStartIdx = 0;
  QString sKey, sValue;
  while ( ParseKeyValue(sConfig, sKey, sValue, nStartIdx) ) {
      if ( "size" == sKey )
      nSize = sValue.toInt();
    if ( "icon" == sKey )
      Icons << sValue;
    if ( "click" == sKey )
      Clicks << sValue;
  }

  IconStack* pIconStack = new IconStack( m_pYasp );
  for ( int i=0; i<Icons.size(); ++i ) {
    QPixmap p = KIconLoader::global()->loadIcon(Icons[i], KIconLoader::NoGroup, nSize);
    pIconStack->AddIcon( p, Clicks.value(i, "") );
  }
  m_pYasp->layout()->addItem( pIconStack, m_Row, 0, 1, 2 );
  m_Row++;

  return true;
}


Qt::Alignment ScriptParser::StringToAlignment( const QString& s )
{
  return s.toLower()=="left" ? Qt::AlignLeft : s.toLower()=="right" ? Qt::AlignRight : Qt::AlignCenter;
}


bool ScriptParser::WaitForDependencies( const QStringList& Dependencies )
{
  if ( m_bParsingSensors ) {
    while ( m_pYasp->DependenciesFulfilled(Dependencies) == false ) {
      if ( m_bReparseRequested )
        return false;
      else
        msleep( 50 );
    }
    return true;
  }
  else
    return m_pYasp->DependenciesFulfilled( Dependencies );
}


QString& ScriptParser::ReplaceFormatChars( QString& sIn ) const
{
  return sIn.replace( "\\n", "\n" ).replace( "\\t", "\t" );
}

#include "scriptparser.moc"
