/***************************************************************************
 *   Copyright (C) 2009 by Andreas Fink <finkandreas@web.de>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// Kde includes
#include <Plasma/Meter>
#include <Plasma/SignalPlotter>
#include <KConfigDialog>
#include <KProcess>


// Qt includes
#include <QGraphicsGridLayout>
#include <QMenu>
#include <QGraphicsSceneContextMenuEvent>
#include <QFile>
#include <QTextStream>
#include <QApplication>
#include <QFileDialog>
#include <QTimer>
#include <QTextCodec>

#include <unistd.h>

// local includes
#include "yasp-scripted.h"
#include "scriptparser.h"
#include "frame.h"
#include "label.h"



// This is the command that links your applet to the .desktop file
K_EXPORT_PLASMA_APPLET(yasp_scripted, Yasp)
 
 
Yasp::Yasp(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args), m_pLayout(0), m_pScriptParser(0), m_bInitted(false)
{
}


Yasp::~Yasp()
{
  if ( m_bInitted ) {
    Clear();
    delete m_pScriptParser;
  }
}


void Yasp::init()
{
  setHasConfigurationInterface(true);
  setAspectRatioMode( Plasma::IgnoreAspectRatio );
  //resize( 50, 50 );
  QString sHome = getenv("HOME");
  m_sScriptPath = sHome + "/.yasp.script";
  KConfigGroup cg = config();
  m_sScriptPath = cg.readEntry( "scriptpath", m_sScriptPath );

  m_pScriptParser = new ScriptParser( this );
  qRegisterMetaType<SourceIntervalPair>("SourceIntervalPair");
  qRegisterMetaType<SensorCmdPair>("SensorCmdPair");
  qRegisterMetaType<SensorLabelPair>("SensorLabelPair");
  qRegisterMetaType<SensorMeterPair>("SensorMeterPair");
  qRegisterMetaType<SensorPlotterPair>("SensorPlotterPair");
  qRegisterMetaType<SensorTextPair>("SensorTextPair");
  connect( m_pScriptParser, SIGNAL(EngineSensorAdded(SourceIntervalPair,SensorCmdPair)),
           this, SLOT(AddEngineSensor(SourceIntervalPair,SensorCmdPair)) );
  connect( m_pScriptParser, SIGNAL(ProgramSensorAdded(SensorCmdPair,int)),
           this, SLOT(AddProgramSensor(SensorCmdPair,int)) );
  connect( m_pScriptParser, SIGNAL(LabelAdded(SensorLabelPair,int)),
           this, SLOT(AddLabel(SensorLabelPair,int)) );
  connect( m_pScriptParser, SIGNAL(MeterAdded(SensorMeterPair,int)),
           this, SLOT(AddMeter(SensorMeterPair,int)) );
  connect( m_pScriptParser, SIGNAL(PlotterAdded(SensorPlotterPair,int)),
           this, SLOT(AddPlotter(SensorPlotterPair,int)) );
  connect( m_pScriptParser, SIGNAL(TextAdded(SensorTextPair,int)),
           this, SLOT(AddText(SensorTextPair,int)) );
  connect( this, SIGNAL(WaitingForDependency(QString)), this, SLOT(ShowWaitingDependency(QString)) );
  for (int i=0; i<PROCESS_COUNT; ++i) {
    connect( &m_Process[i], SIGNAL(finished(int, QProcess::ExitStatus)), &m_ProcessSignalMapper, SLOT(map()) );
    m_ProcessSignalMapper.setMapping( &m_Process[i], i );
  }
  connect( &m_ProcessSignalMapper, SIGNAL(mapped(int)), this, SLOT(ProcessFinished(int)) );
  QTextCodec::setCodecForCStrings( QTextCodec::codecForLocale() );
  ParseConfig();
  m_bInitted = true;
}


void Yasp::sourceAdded( QString sSource )
{
//  qDebug() << "SourceAdded:" << sSource << m_ScheduledSources;
  if ( m_ScheduledSources.contains(sSource) ) {
    FakeConnect( sSource );
    m_ScheduledSources.remove( sSource );
  }
//  qDebug() << "Source successfully added";
}


void Yasp::FakeConnect( const QString& sourceName )
{
//  qDebug() << "Fake Connecting";
  if ( ! m_SourceSpecs.contains(sourceName) )
    return;
  const SourceIntervalPair& sip = m_SourceSpecs.value( sourceName );
  Plasma::DataEngine* pEngine = dataEngine( sip.Engine );
  if ( ! pEngine->sources().contains(sip.Source) ) {
    if ( ! m_ConnectedEngines.contains(sip.Engine) ) {
      m_ConnectedEngines << sip.Engine;
      connect( pEngine, SIGNAL(sourceAdded(QString)), this, SLOT(sourceAdded(QString)) );
    }
    m_ScheduledSources.insert( sip.Source );
  }

  m_ScheduledReconnectSources.insert( sip.Source );
  pEngine->connectSource( sip.Source, this, 20 );
}


void Yasp::RealConnect( const QString& sourceName )
{
  if ( ! m_SourceSpecs.contains(sourceName) )
    return;

  m_ScheduledReconnectSources.remove( sourceName );
  int nInterval = 0;
  Plasma::DataEngine* pEngine = dataEngine( m_SourceSpecs.find(sourceName).value().Engine );
  QHash<QString, SourceIntervalPair>::iterator it = m_SourceSpecs.find( sourceName );
  while ( it != m_SourceSpecs.end() && it.key() == sourceName ) {
    const SourceIntervalPair& sip = it.value();
    if ( sip.Interval != 0  ) {
      if ( nInterval==0 || sip.Interval<nInterval ) {
        nInterval = sip.Interval;
      }
      ++it;
    }
    else {
      QHash<QString, SensorCmdPair>::iterator it2 = m_EngineKeys.find( sourceName );
      while ( it2 != m_EngineKeys.end() && it2.key() == sourceName ) {
        if ( it2.value().Cmd == sip.Key )
          it2 = m_EngineKeys.erase( it2 );
        else
          ++it2;
      }
      it = m_SourceSpecs.erase( it );
    }
  }

  if ( nInterval > 0 ) {
    pEngine->connectSource( sourceName, this, nInterval );
  }
  else {
    pEngine->disconnectSource( sourceName, this );
  }
}


void Yasp::dataUpdated( const QString& sourceName, const Plasma::DataEngine::Data& data )
{
//  qDebug() << "DataUpdated..." << sourceName << data ;
  QMutexLocker locker( &m_Mutex );
  if ( ! m_EngineKeys.contains(sourceName) ) {
    return;
  }

  bool bAllKeysAvailable = true;
  QHash<QString, SensorCmdPair>::const_iterator it = m_EngineKeys.find( sourceName );
  while ( it != m_EngineKeys.end() && it.key() == sourceName ) {
    const SensorCmdPair& scp = it.value();
    if ( data.contains(scp.Cmd) )
      m_Sensors[scp.Sensor] = MathOperation( data[scp.Cmd], scp.MathOperation, scp.UsedSensors );
    else
      bAllKeysAvailable = false;
    ++it;
  }

  if ( bAllKeysAvailable &&  m_ScheduledReconnectSources.contains(sourceName) ) {
    // this code is only executed when we need to reconnect from 20ms to the desired interval
    RealConnect( sourceName );
  }
}


QList< QAction* > Yasp::contextualActions()
{
  QAction* p1 = new QAction( "Reparse script", this );
  connect( p1, SIGNAL(triggered()), this, SLOT(ParseConfig()) );
  return QList<QAction*>() << p1;
}


void Yasp::createConfigurationInterface( KConfigDialog* pParent )
{
  QWidget *widget = new QWidget();
  ui.setupUi(widget);
  ui.ScriptPathLineEdit->setText( m_sScriptPath );
  connect( ui.SearchScriptPathButton, SIGNAL(clicked()), this, SLOT(OpenFileDialog()) );
  if ( backgroundHints() == Plasma::Applet::NoBackground )
    ui.NoBackgroundHint->setChecked( true );
  else if ( backgroundHints() == Plasma::Applet::DefaultBackground )
    ui.DefaultBackgroundHint->setChecked( true );
  else if ( backgroundHints() == Plasma::Applet::TranslucentBackground )
    ui.TranslucentBackgroundHint->setChecked( true );
  m_pConfig = pParent;
  pParent->addPage( widget, i18n("Configuration"), "Yasp" );
  connect( pParent, SIGNAL(applyClicked()), this, SLOT(ConfigAccepted()) );
  connect( pParent, SIGNAL(okClicked()), this, SLOT(ConfigAccepted()) );
}


void Yasp::OpenFileDialog()
{
  QString s = QFileDialog::getOpenFileName( m_pConfig, "Choose new script file" );
  if ( ! s.isNull() )
    ui.ScriptPathLineEdit->setText( s );
}


void Yasp::ConfigAccepted()
{
  KConfigGroup cg = config();
  if ( m_sScriptPath != ui.ScriptPathLineEdit->text() ) {
    m_sScriptPath = ui.ScriptPathLineEdit->text();
    cg.writeEntry( "scriptpath", m_sScriptPath );
    ParseConfig();
  }
  if ( ui.NoBackgroundHint->isChecked() )
    setBackgroundHints( Plasma::Applet::NoBackground );
  else if ( ui.DefaultBackgroundHint->isChecked() )
    setBackgroundHints( Plasma::Applet::DefaultBackground );
  else if ( ui.TranslucentBackgroundHint->isChecked() )
    setBackgroundHints( Plasma::Applet::TranslucentBackground );
  cg.writeEntry( "backgroundhint", (int)backgroundHints() );

  emit configNeedsSaving();
}


void Yasp::constraintsEvent( Plasma::Constraints /*constraints*/ )
{
  KConfigGroup cg = config();
  setBackgroundHints( (Plasma::Applet::BackgroundHint)cg.readEntry("backgroundhint", 1) );
}


void Yasp::timerEvent( QTimerEvent* pEvent )
{
  if ( m_ProgramKeys.contains(pEvent->timerId()) )
    TimeoutUpdateSensor( m_ProgramKeys[pEvent->timerId()] );
  else if ( m_Values.contains(pEvent->timerId()) )
    TimeoutUpdateLabel( m_Values[pEvent->timerId()] );
  else if ( m_Meters.contains(pEvent->timerId()) )
    TimeoutUpdateMeter( m_Meters[pEvent->timerId()] );
  else if ( m_Plotters.contains(pEvent->timerId()) )
    TimeoutUpdatePlotter( m_Plotters[pEvent->timerId()] );
  else if ( m_Texts.contains(pEvent->timerId()) )
    TimeoutUpdateText( m_Texts[pEvent->timerId()] );
  else Plasma::Applet::timerEvent( pEvent );
}


void Yasp::TimeoutUpdateSensor( const SensorCmdPair& scp )
{
  int nFreeId = -1;
  for ( int i=0; i<PROCESS_COUNT; ++i ) {
    if ( m_ProcessSensors[i].Sensor.isEmpty() ) {
      nFreeId = i;
      break;
    }
  }

  if ( nFreeId == -1 ) {
    m_ProcessQueue.enqueue(scp);
  }
  else {
//    qDebug() << scp.Sensor << nFreeId;
    m_ProcessSensors[nFreeId] = scp;
    if ( scp.Cmd.isEmpty() ) {
      ProcessFinished( nFreeId );
    }
    else {
      m_Process[nFreeId].setShellCommand( scp.Cmd );
      m_Process[nFreeId].setOutputChannelMode( KProcess::MergedChannels );
      m_Process[nFreeId].start();
    }
  }
}


void Yasp::ProcessFinished( int nId )
{
  const SensorCmdPair& scp = m_ProcessSensors[nId];
  QString s;
  if ( !scp.Cmd.isEmpty() )
    s = m_Process[nId].readAllStandardOutput().trimmed();
  m_Sensors[ scp.Sensor ] = MathOperation( s, scp.MathOperation, scp.UsedSensors );
  SensorCmdPair d;
  m_ProcessSensors[nId] = d;
  if ( !m_ProcessQueue.isEmpty() )
    TimeoutUpdateSensor( m_ProcessQueue.dequeue() );
}


void Yasp::TimeoutUpdateLabel( const SensorLabelPair& slp )
{
  slp.pDisplayLabel->setText( PasteSensorValues(slp.Format, slp.UsedSensors) );
}


void Yasp::TimeoutUpdateMeter( const SensorMeterPair& smp )
{
  Plasma::Meter* m = smp.pMeter;
  m->setMinimum( PasteSensorValues(smp.Min, smp.UsedSensors).toDouble() );
  m->setMaximum( PasteSensorValues(smp.Max, smp.UsedSensors).toDouble() );
  m->setValue( PasteSensorValues(smp.Value, smp.UsedSensors).toDouble() );
  for ( QHash<int, QString>::const_iterator it = smp.Labels.begin(); it != smp.Labels.end(); ++it )
    m->setLabel( it.key(), PasteSensorValues(it.value(), smp.UsedSensors) );
}


void Yasp::TimeoutUpdatePlotter( const SensorPlotterPair& spp )
{
  Plasma::SignalPlotter* p = spp.pPlotter;
  QList<double> samples;
  for ( int i=0; i<spp.Plots.size(); ++i )
    samples << PasteSensorValues(spp.Plots[i], spp.UsedSensors).toDouble();
  p->addSample( samples );
}


void Yasp::TimeoutUpdateText( const SensorTextPair& stp )
{
  stp.pTextLabel->setText( PasteSensorValues(stp.Format, stp.UsedSensors) );
}


QString Yasp::PasteSensorValues( const QString& sFormat, const QStringList& UsedSensors )
{
  QString sResult( sFormat );
  int i=1;
  foreach( QString sSensor, UsedSensors ) {
    QString sAfter;
    if ( m_Sensors[sSensor].type() == QVariant::Double )
      sAfter = QString("%1").arg(m_Sensors[sSensor].toDouble(), 0, 'f', 2);
    else
      sAfter = m_Sensors[sSensor].toString();
    QString sBefore = QString("$%1").arg(i++);
    sResult.replace( sBefore, sAfter );
  }
  return sResult;
}


bool Yasp::DependenciesFulfilled( const QStringList& Dependencies )
{
  foreach( QString d, Dependencies )
    if ( !m_Sensors.contains(d) ) {
      emit WaitingForDependency( d );
      return false;
    }
  return true;
}


QStringList Yasp::BrokenDependencies( const QStringList& Dependencies )
{
  QStringList result;
  foreach ( QString d, Dependencies )
    if ( !m_Sensors.contains(d) )
      result << d;
  return result;
}


void Yasp::Clear()
{
  QMutexLocker locker( &m_Mutex );
  if ( m_pScriptParser ) {
    disconnect( m_pScriptParser, SIGNAL(finished()), this, SLOT(ParsingFinished()) );
    m_pScriptParser->ReparseRequested();
    m_pScriptParser->wait();
  }
  for ( int i=0; i<PROCESS_COUNT; ++i ) {
    m_Process[i].waitForFinished(-1);
  }
  QHash<QString, SourceIntervalPair>::const_iterator it = m_SourceSpecs.begin();
  while ( it != m_SourceSpecs.end() ) {
    const SourceIntervalPair& sip = it.value();
//    qDebug() << "Disconnecting from engine" << sip.Engine << "the source" << sip.Source;
    dataEngine(sip.Engine)->disconnectSource(sip.Source, this);
//    qDebug() << "Successfully disconnected";
    ++it;
  }
  foreach( int timerid, m_ProgramKeys.keys() )
    killTimer( timerid );
  foreach( int timerid, m_Values.keys() )
    killTimer( timerid );
  foreach( int timerid, m_Meters.keys() )
    killTimer( timerid );
  foreach( int timerid, m_Plotters.keys() )
    killTimer( timerid );
  foreach( int timerid, m_Texts.keys() )
    killTimer( timerid );
  m_ScheduledSources.clear();
  m_SourceSpecs.clear();
  m_ScheduledReconnectSources.clear();
  m_EngineKeys.clear();
  m_ProgramKeys.clear();
  m_Sensors.clear();
  m_Values.clear();
  m_Meters.clear();
  m_Plotters.clear();
  m_Texts.clear();
  while ( childItems().empty() == false )
    delete childItems().first();
  setLayout( 0 );
  connect( m_pScriptParser, SIGNAL(finished()), this, SLOT(ParsingFinished()) );
}


void Yasp::ParseConfig()
{
  m_bParsingError = false;
  Clear();
  m_pLayout = new QGraphicsGridLayout;
  m_pInfoLabel = new Label( QPixmap(), this );
  m_pLayout->addItem( m_pInfoLabel, 0, 0, 1, 1 );
  setLayout( m_pLayout );
  m_pScriptParser->start();
}


void Yasp::ParsingError( const QString& sErrorMessage )
{
  Clear();
  m_bParsingError = true;
  m_pLayout = new QGraphicsGridLayout;
  Label* l = new Label( QPixmap(), this );
  l->setText( sErrorMessage );
  m_pLayout->addItem( l, 0, 0, 1, 1 );
  setLayout( m_pLayout );
}


void Yasp::ParsingFinished()
{
  if ( m_bParsingError )
    return;

  m_pLayout = new QGraphicsGridLayout;
  m_pScriptParser->ParseGuiElements();
  delete m_pInfoLabel;
  m_pInfoLabel = 0;
  m_pLayout->setContentsMargins( 0, 0, 0, 0 );
  m_pLayout->setSpacing( 0 );
  setLayout( m_pLayout );
  setMinimumSize( m_pLayout->minimumSize() + (boundingRect().size()-this->contentsRect().size()) );
  update();
}


void Yasp::AddEngineSensor( const SourceIntervalPair& sip, const SensorCmdPair& scp )
{
  m_EngineKeys.insertMulti( sip.Source, scp );
  m_SourceSpecs.insertMulti( sip.Source, sip );
  FakeConnect( sip.Source );
}


void Yasp::AddProgramSensor( const SensorCmdPair& scp, int nUpdateInterval )
{
//  qDebug() << __FILE__ << __LINE__ << scp.Sensor;
  TimeoutUpdateSensor( scp );  // update the first time
  if ( nUpdateInterval != 0 )  // register for periodic updates
    m_ProgramKeys[ startTimer(nUpdateInterval) ] = scp;
}


void Yasp::AddLabel( const SensorLabelPair& slp, int nUpdateInterval )
{
  TimeoutUpdateLabel( slp );
  if ( 0 != nUpdateInterval )
    m_Values[ startTimer(nUpdateInterval) ] = slp;
}


void Yasp::AddMeter( const SensorMeterPair& smp, int nUpdateInterval )
{
  TimeoutUpdateMeter( smp );
  if ( 0 != nUpdateInterval )
    m_Meters[ startTimer(nUpdateInterval) ] = smp;
}


void Yasp::AddPlotter( const SensorPlotterPair& spp, int nUpdateInterval )
{
  TimeoutUpdatePlotter( spp );
  if ( 0 != nUpdateInterval )
    m_Plotters[ startTimer(nUpdateInterval) ] = spp;
}


void Yasp::AddText( const SensorTextPair& stp, int nUpdateInterval )
{
  TimeoutUpdateText( stp );
  if ( 0 != nUpdateInterval )
    m_Texts[ startTimer(nUpdateInterval) ] = stp;
}


void Yasp::ShowWaitingDependency( const QString& Dependency )
{
  m_pInfoLabel->setText( "Waiting for: " + Dependency );
}


QVariant Yasp::MathOperation( const QVariant& Value, const QString& sMathOperation, const QStringList& UsedSensors )
{
  QVariant v(Value);
  if ( sMathOperation.section(' ', 0, 0).trimmed() == "int" )
    v = MathOperation<int>( v, sMathOperation.section(' ', 1).trimmed(), UsedSensors );
  if ( sMathOperation.section(' ', 0, 0).trimmed() == "double" )
    v = MathOperation<double>( v, sMathOperation.section(' ', 1).trimmed(), UsedSensors );
  return v;
}

#include "yasp-scripted.moc"
