/***************************************************************************
 *   Copyright (C) 2009 by Andreas Fink <finkandreas@web.de>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// Here we avoid loading the header multiple times
#ifndef YASP_SCRIPTED_H
#define YASP_SCRIPTED_H

#include <Plasma/Applet>

#include <KProcess>
#include <QHash>
#include <QMultiHash>
#include <QStack>
#include <QRegExp>
#include <QMutex>
#include <QSignalMapper>
#include <QQueue>

#include "frame.h"
#include "ui_config.h"
#include "structs.h"

class Label;
class ScriptParser;
class QGraphicsGridLayout;

// Define our plasma Applet
class Yasp : public Plasma::Applet
{
  Q_OBJECT

public:

  // Basic Create/Destroy
  Yasp(QObject *parent, const QVariantList &args);
  virtual ~Yasp();

  void init();
  virtual QList< QAction* > contextualActions();
  virtual void createConfigurationInterface(KConfigDialog *parent);
  bool DependenciesFulfilled( const QStringList& Dependencies );
  QStringList BrokenDependencies( const QStringList& Dependencies );
  void ParsingError( const QString& sErrorLine );
  QGraphicsGridLayout* layout() { return m_pLayout; }
  QString ScriptPath() { return m_sScriptPath; }


public slots:
  void sourceAdded( QString sSource );
  void dataUpdated( const QString& sourceName, const Plasma::DataEngine::Data& data );
  void AddEngineSensor( const SourceIntervalPair& sip, const SensorCmdPair& scp );
  void AddProgramSensor( const SensorCmdPair& scp, int nUpdateInterval );
  void AddLabel( const SensorLabelPair& slp, int nUpdateInterval );
  void AddMeter( const SensorMeterPair& smp, int nUpdateInterval );
  void AddPlotter( const SensorPlotterPair& spp, int nUpdateInterval );
  void AddText( const SensorTextPair& stp, int nUpdateInterval );
  void ProcessFinished( int nId );
  void ShowWaitingDependency( const QString& Dependency );

protected slots:
  void ParseConfig();
  void ParsingFinished();
  void ConfigAccepted();
  void OpenFileDialog();

signals:
  void WaitingForDependency( const QString& Dependency );

protected:
  virtual void constraintsEvent( Plasma::Constraints constraints );
  virtual void timerEvent ( QTimerEvent * event );
  void FakeConnect( const QString& sourceName );
  void RealConnect( const QString& sourceName );
  void TimeoutUpdateSensor( const SensorCmdPair& scp );
  void TimeoutUpdateLabel( const SensorLabelPair& slp );
  void TimeoutUpdateMeter( const SensorMeterPair& smp );
  void TimeoutUpdatePlotter( const SensorPlotterPair& spp );
  void TimeoutUpdateText( const SensorTextPair& stp );
  QString PasteSensorValues( const QString& sFormat, const QStringList& UsedSensors );
  void Clear();
  QVariant MathOperation( const QVariant& Value, const QString& sMathOperation, const QStringList& UsedSensors );

#define PROCESS_COUNT 5
  KProcess m_Process[PROCESS_COUNT];
  SensorCmdPair m_ProcessSensors[PROCESS_COUNT];
  QQueue<SensorCmdPair> m_ProcessQueue;
  QSignalMapper m_ProcessSignalMapper;
  QGraphicsGridLayout* m_pLayout;
  QString m_sScriptPath;
  Ui::config ui;
  QWidget* m_pConfig;
  ScriptParser* m_pScriptParser;
  bool m_bParsingError;
  QMutex m_Mutex;
  Label* m_pInfoLabel;
  bool m_bInitted;

  /// A hash of scheduled sources for a dataEngine.
  QSet< QString > m_ScheduledSources;
  QSet< QString > m_ScheduledReconnectSources;
  QMultiHash< QString, SourceIntervalPair > m_SourceSpecs;

  /// Set of all engines for which the sourceAdded signal was already connected
  QSet<QString> m_ConnectedEngines;

  /// contains for every Plasma::DataEngine::source (the QString key) the SensorCmdPair which needs to be updated
  QMultiHash< QString, SensorCmdPair > m_EngineKeys;

  /// contains for timer-id (the int key) a SensorCmdPair, which needs to be updated
  QHash< int, SensorCmdPair > m_ProgramKeys;

  /// contains for every defined sensor (the QString key), a QVariant value
  QHash< QString, QVariant > m_Sensors;

  /// contains for timer-id (the int key) the SensorLabelPair which needs to be updated
  QHash< int, SensorLabelPair > m_Values;
  /// contains for a timer-id (the int key) the SensorMeterPair which needs to be updated
  QHash< int, SensorMeterPair > m_Meters;
  /// contains for a timer id (the int key) the SensorPlotterPair which needs  to be updated
  QHash< int, SensorPlotterPair > m_Plotters;
  QHash< int, SensorTextPair > m_Texts;

  template <class T> QVariant MathOperation( const QVariant& Value, const QString& sOperation, const QStringList UsedSensors )
  {
    QRegExp DoubleRegExp( "[0-9]*([.,][0-9]*)?" );
    QRegExp OperatorRegExp( "[*/+-]" );
    QStack<T> Stack;
    QStringList sOperands = sOperation.split(' ', QString::SkipEmptyParts );
    foreach( QString sOperand, sOperands ) {
      if ( sOperand.startsWith('$') ) {
        int nArg = sOperand.mid(1, 10).toInt();
        if ( nArg == 0 )
          Stack.push( (T)Value.toDouble() );
        else if ( nArg-1 < UsedSensors.size() )
          Stack.push( (T)m_Sensors[UsedSensors[nArg-1]].toDouble() );
        else
          return Value;   // something went wrong
      }
      else if ( DoubleRegExp.exactMatch(sOperand.trimmed()) )
        Stack.push( (T)sOperand.toDouble() );  // casts to an int if T==int
      else if ( OperatorRegExp.exactMatch(sOperand.trimmed()) && Stack.size() >= 2 ) {
        T Op1 = Stack.pop();
        T Op2 = Stack.pop();
        if ( sOperand.trimmed() == "+" )
          Stack.push( Op1 + Op2 );
        else if ( sOperand.trimmed() == "-" )
          Stack.push( Op2 - Op1 );
        else if ( sOperand.trimmed() == "*" )
          Stack.push( Op2 * Op1 );
        else if ( sOperand.trimmed() == "/" ) {
          if ( Op1 == 0 )
            Stack.push( 0 );
          else
            Stack.push( Op2 / Op1 );
        }
      }
      else
        return Value;   // something went wrong
    }
    if ( Stack.size() == 1 )
      return QVariant( Stack.pop() );
    else
      return Value;   // something went wrong
  }
};
#endif
